import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import netscape.ldap.*; 
import javax.servlet.*;

public class LDAPServlet extends GenericServlet implements Runnable
{
  
public final static int SAVE      = 1;
public final static int OPEN      = 2;
public final static int DELETE    = 3;
public final static int COPY      = 4;
public final static int PASTE     = 5;
public final static int CUT       = 6;
public final static int NEWFILE   = 7;
public final static int NEWFOLDER = 8;
public final static int RENAME    = 9;
public final static int GETTREE   = 10;
  
 
public String centralLDAPHost;
public int    centralLDAPPort;  
public int WFEPort; 
public Vector hosts,ports,conns;
public ServerSocket WFESock;
public Thread socketThread;
public int counter=0;
public int socketCounter=0; 
public Hashtable LDAPHostForSite;
public Vector Machines;
  
public Socket []socket_list = new Socket [16];
public ObjectOutputStream []oos_list = new ObjectOutputStream[16];
public ObjectInputStream []ois_list = new ObjectInputStream[16];
public DataInputStream []dis_list = new DataInputStream[16];  


  //Params will be
  // WFEPort=xxxx&localhost=x.y.z&NCSA=modi4.ncsa.uiuc.edu&
  //NPAC=osprey4.npac.syr.edu
public  void getSites(ServletConfig Params)
  {
    String name,value;

    LDAPHostForSite = new Hashtable();
    Enumeration paramNames =  Params.getInitParameterNames();
    while(paramNames.hasMoreElements()){
      name = (String)paramNames.nextElement();
      value = Params.getInitParameter(name);
      LDAPHostForSite.put(name,value);
    }
    value   = (String)LDAPHostForSite.get("localhost");
    int colon = value.indexOf(":");
    centralLDAPHost = value.substring(0,colon);
    centralLDAPPort = Integer.parseInt(value.substring(colon+1));
    WFEPort         = Integer.parseInt((String)LDAPHostForSite.get("WFEPort"));
    
  }
  

public void init(ServletConfig Params) throws ServletException
   {
     try {
       super.init(Params);
     } catch (ServletException e)
       {
	 System.out.println(" Exception in init() of  LDAPServlet"+e);
       }
      getSites(Params);
      
      //LDAP server hosts & ports and their connection handles
      hosts = new Vector();
      ports = new Vector();
      conns = new Vector();
      //Parallel machines. we keep just organize user program-generated files
      Machines = new Vector();
    
      addLDAPServer(centralLDAPHost,centralLDAPPort);
      try{
	WFESock = new ServerSocket(WFEPort);
	WFESock.setSoTimeout(1);
      }catch(IOException e){
	System.out.println("LDAPServlet could not open port:"+WFEPort+":"+e);
      }
      for(int k = 0;k<16;k++){
	socket_list[k]=null;
	oos_list[k] = null;
	ois_list[k] = null;
      }
     
      Machines.addElement("Origin2000");
      Machines.addElement("Convex");
      Machines.addElement("SunSolaris");
      Machines.addElement("PowerChallenge");
      //start thread
      socketThread = new Thread(this);
      socketThread.setDaemon(true);
      socketThread.start();
    
   }

public void service(ServletRequest req, ServletResponse res)
    throws ServletException, IOException
  {
     counter++;
 
     res.setContentType("text/plain");
     OutputStream os = res.getOutputStream();
     PrintStream out = new PrintStream( os );
     out.println("Host:"+centralLDAPHost);
     out.println("Port:"+centralLDAPPort);
     out.println("WFEPort:"+WFEPort);
     out.println("LDAPServlet:"+counter);
     out.flush();
   }

  public void   run() 
  {
    boolean flag = true;
    
    while(flag) {
      try{
	Socket s = WFESock.accept();
	oos_list[socketCounter] = new ObjectOutputStream(s.getOutputStream());
	ois_list[socketCounter] = new ObjectInputStream(s.getInputStream());
	dis_list[socketCounter] = new DataInputStream(s.getInputStream());
	System.out.println("client connected !!"+oos_list[socketCounter]+
			   "&&"+ois_list[socketCounter]);

	socket_list[socketCounter++]  = s;
      }catch(InterruptedIOException e){
       //System.out.println("InterruptedIOException during accepting clinet ");
      }catch(IOException e){
	System.out.println("IOException during accepting clinet ");
      }
   
      for(int k=0;k < socketCounter; k++){
	try{
	  if ((dis_list[k] != null) && (dis_list[k].available() > 0))
	    {
	      processWFECommand(k);
	    }
	  }catch ( IOException e ) 
	    {
	      System.out.println( "IOException(run): " + e.toString() );
	    }
      }
           
    }

    
  }
  
  
  
  
  //add this LDAP Server and make connection to it
public void addLDAPServer(String host, int port)
  {
    LDAPConnection oneConn = null;
    
    try{
      oneConn = new LDAPConnection();
      oneConn.connect(host,port);
    }catch ( LDAPException e ) {
      System.out.println( "Unable to connect: " + e.toString() );
    }  
    hosts.addElement(host);
    ports.addElement(new Integer(port));
    conns.addElement(oneConn);
  }
  

public void removeLDAPServer(String host, int port)
  {
    int index = hosts.indexOf(host);
    LDAPConnection ld = (LDAPConnection)(conns.elementAt(index));
    
    if ( (ld != null) && ld.isConnected() ) {
      try {
	ld.disconnect();
      } catch ( LDAPException e ) {
	System.out.println( "Unable disconnect: " + e.toString() );
      }
    }
    
    conns.removeElement(ld);
    hosts.removeElement(host);
    ports.removeElement(new Integer(port));
      
  }

public LDAPConnection connection(String host,int port)
  {
    LDAPConnection ld =null;
    
    int index = hosts.indexOf(host);
    if (index < 0){
      addLDAPServer (host,port);
      index = hosts.indexOf(host);
    }
    ld = (LDAPConnection)(conns.elementAt(index));
        
    return ld;
  }

public Hashtable readEntry(String LDAPHost,int port,String entryRDN)
  {
    /* Read all attributes */
    Hashtable attributes = new  Hashtable();
    
    try{
      LDAPConnection ld = connection(LDAPHost,port);
      LDAPEntry findEntry = ld.read( entryRDN );
  
      /* Get the attributes of the entry */
      LDAPAttributeSet findAttrs = findEntry.getAttributeSet();
      Enumeration enumAttrs = findAttrs.getAttributes();

      System.out.println("reading entries of RDN="+ findEntry.getDN() );
      /* Loop on attributes */
      while ( enumAttrs.hasMoreElements() ) {
	LDAPAttribute anAttr =
	  (LDAPAttribute)enumAttrs.nextElement();
	String attrName = anAttr.getName();
	/* Loop on values for this attribute */
	Enumeration enumVals = anAttr.getStringValues();
	Vector attVals = new Vector();
	while(enumVals.hasMoreElements())
	  attVals.addElement(enumVals.nextElement());
	attributes.put(attrName,attVals);
      }
      
      }catch( LDAPException e ) 
      {
	System.out.println( "Error during getting attributes:"+e.toString());
      }
      return attributes;  
      
  }
    
public String createRDN(Hashtable enumAttrs)
  {
    String filename = (String)((Vector)enumAttrs.get("filename")).elementAt(0); 
    String cn       = (String)((Vector)enumAttrs.get("cn")).elementAt(0); 
    String hn       = (String)((Vector)enumAttrs.get("machine")).elementAt(0); 
    String o        = (String)((Vector)enumAttrs.get("organization")).elementAt(0); 

    String RDN = "";
    RDN += "cn="+filename;
    RDN += ",cn="+cn;
    RDN += ",hn="+hn;
    RDN += ",o="+o;
    RDN += ",o=WebFlow,c=US";
    System.out.println("Constructed RDN:"+RDN);
    return RDN;
    
  }
 
public String getFileFromEntry(Hashtable enumAttrs)
  {
    
    String o = (String)((Vector)enumAttrs.get("organization")).elementAt(0);
    String    remoteLDAPHost = (String)LDAPHostForSite.get(o);
    if (remoteLDAPHost.equalsIgnoreCase("localhost")){
      System.out.println("Sending local file");
      String fileContent = (String)((Vector)enumAttrs.
				    get("content")).elementAt(0);
      return fileContent ;
    }
    else{
      int index = remoteLDAPHost.indexOf(':');
      int remoteLDAPPort =Integer.parseInt(remoteLDAPHost.substring(index+1));
      remoteLDAPHost = remoteLDAPHost.substring(0,index);
      String remoteRDN = createRDN(enumAttrs);
      enumAttrs = readEntry( remoteLDAPHost,remoteLDAPPort,remoteRDN) ;
      String fileContent = (String)((Vector)enumAttrs.get("content")).elementAt(0);
      return fileContent ;
    }
    
  }
 
public void sendWFECommandResult(ObjectOutputStream oos,Object result)
  {
    System.out.println("LDAPServlet send result:");
    
    try{
      oos.writeObject(result);
    }catch(OptionalDataException e){
      System.out.println("sendWFECommand "+e);
    }catch(IOException e){
      System.out.println("sendWFECommand "+e);
    }
    
  }



public void sendInt(ObjectOutputStream oos,int data)
  {
    try{
      oos.writeInt(data);
    }catch(IOException e){
      System.out.println("sendInt "+e);
    }
  }

public String[] getSites()
  {
    Vector v = new Vector();
    Enumeration enum = LDAPHostForSite.keys();
    while(enum.hasMoreElements())
      v.addElement(enum.nextElement());
    
    String []vals = new String[v.size()];
    for(int k=0;k<v.size();k++)
      vals[k] = (String)v.elementAt(k);
    return vals;
  }
  
public String[] getMachines()
  {
    int len = Machines.size();
    String []vals = new String[len];
    for(int k=0;k<len;k++)
      vals[k] = (String)Machines.elementAt(k);
    return vals;
    
  }
  

 public void searchFullandSend (ObjectOutputStream oos,
				String searchBase,String searchFilter)
  {
    int r;
    
    /* Read all attributes */
    try {
      LDAPConnection ld = connection(centralLDAPHost,centralLDAPPort);

      LDAPSearchConstraints cons = ld.getSearchConstraints();
      cons.setBatchSize( 1 );
      LDAPSearchResults res = ld.search( searchBase, LDAPConnection.SCOPE_SUB,
		       searchFilter, null, false,  cons );

      /* Loop on results until finished */
      sendInt(oos,res.hasMoreElements()?1:0);
      while ( res.hasMoreElements() ) {
	/* Next directory entry */
	LDAPEntry findEntry = (LDAPEntry)res.nextElement();
	System.out.println( findEntry.getDN() );
	String newRDN = findEntry.getDN();
	sendWFECommandResult(oos,newRDN);
	
	/* Get the attributes of the entry */
	LDAPAttributeSet findAttrs = findEntry.getAttributeSet();
	Enumeration enumAttrs = findAttrs.getAttributes();
	/* Loop on attributes */
	sendInt(oos,enumAttrs.hasMoreElements()?1:0);
	while ( enumAttrs.hasMoreElements() ) {
	  LDAPAttribute anAttr =
	    (LDAPAttribute)enumAttrs.nextElement();
	  String attrName = anAttr.getName();
	  sendWFECommandResult(oos,attrName);
	  /* Loop on values for this attribute */
	  Enumeration enumVals = anAttr.getStringValues();
	  sendInt(oos,enumVals.hasMoreElements()?1:0);
	  
	  while ( enumVals.hasMoreElements() ) {
	    String aVal = ( String )enumVals.nextElement();
	    sendWFECommandResult(oos,aVal);
	    sendInt(oos,enumVals.hasMoreElements()?1:0);
	  }
	  sendInt(oos,enumAttrs.hasMoreElements()?1:0);
	}
	sendInt(oos,res.hasMoreElements()?1:0);
      }
      //send sites and their LDAP server hosts
      sendWFECommandResult(oos,getSites());
      sendWFECommandResult(oos,getMachines());
      oos.flush();
    } catch( LDAPException e ) {
	System.out.println( "SearchFull:Error: " + e.toString() );
      }
    catch(IOException e){
      System.out.println("sendInt "+e);
    }
    System.out.println("We sent all TREE !!!");
    
  }
 
public int numOfColons(String s)
  {
    int ncolons = 0;
    StringTokenizer st = new StringTokenizer(s,",");
    while (st.hasMoreTokens()) 
      ncolons++;
    return ncolons;
}

public String makeNewRDN(String oldDN,String newRDNVal)
  {
    System.out.println("makeNewRDN: oldDN:"+oldDN+"  newRDNVal:"+ newRDNVal);
    
    int equal = oldDN.indexOf("=");
    String    firstAttName = oldDN.substring(0,equal);
    String newRDN = firstAttName+"="+newRDNVal;
    return newRDN;
  }
public String makeNewDN(String oldDN,String newRDN)
  {
    int colon      = oldDN.indexOf(",");
    String rest    = oldDN.substring(colon+1);
    String newEntryDN = newRDN+","+rest;
    return newEntryDN;
  }
   
  //newRDN will be attname=attval 
public void renameEntryX(String LDAPHost,int LDAPPort,String ENTRYDN,
			String newRDN,String []childrenDNs)
  {
    String replaceStr =  newRDN.substring(newRDN.indexOf("=")+1);
    String [][]pairs = findUnmatchedRDNs(ENTRYDN,
					 makeNewDN(ENTRYDN,newRDN),
					 false,
					 replaceStr);
    LDAPConnection ld = connection(LDAPHost,LDAPPort);
    for(int k=0;k<childrenDNs.length;k++)
      updateOneEntry(LDAPHost,LDAPPort,newRDN,
		     childrenDNs[k],pairs);
   
  }

  //newRDN will be attname=attval 
public void renameEntry(String LDAPHost,int LDAPPort,String ENTRYDN,
			String newRDN)
  {
    String replaceStr =  newRDN.substring(newRDN.indexOf("=")+1);
    String [][]pairs = findUnmatchedRDNs(ENTRYDN,
					 makeNewDN(ENTRYDN,newRDN),
					 false,
					 replaceStr);
    LDAPConnection ld = connection(LDAPHost,LDAPPort);
    updateOneEntry(LDAPHost,LDAPPort,newRDN,
		     ENTRYDN,pairs);
  }
  
public String[][] findUnmatchedRDNs(String source,String target,
				    boolean isCopy,String replaceStr)
  {  
    String []attrList = {"filename","codename","projectname",
			 "cn","sim_type","o","","" };
    StringTokenizer s,t;
    Vector  sv=new Vector(),tv=new Vector();
    String  ss,ts;
    int k;
          
    System.out.println("compare :"+source+" with :"+target);
    s = new StringTokenizer(source,",",false);
    t = new StringTokenizer(makeCopyOfDN(target),",",false);
    while((s.hasMoreTokens())&&(t.hasMoreTokens())){
      sv.addElement(s.nextElement());
      tv.addElement(t.nextElement());
    }
    //here you are trying to copy one entry to the place where
    //one doest have same parallel structure as source
    if ((s.hasMoreTokens()&&(!t.hasMoreTokens()))||
	(!s.hasMoreTokens()&&(t.hasMoreTokens())))
      return null;
    
    int offset = attrList.length-sv.size();
    for(k=sv.size()-1;k>=0;k--)
      {
	ss = (String)sv.elementAt(k);
	ts = (String)tv.elementAt(k);
	if (!ss.equalsIgnoreCase(ts))
	  break;
      }
    //If two DN matched completely
    if (k < 0)
      return null;

    //If two DN did not matche
    int len = k+1;
    String [][]pairs = new String[len][3];
    pairs[0][0] = (String)sv.elementAt(0);
    int equal =  pairs[0][0].indexOf("=");
    if (isCopy)//make copy
      pairs[0][1] = pairs[0][0].substring(0,equal+1)+replaceStr;
    else //make cut
      pairs[0][1] = pairs[0][0];
    pairs[0][2] = attrList[offset];
    System.out.println("Replace "+pairs[0][0]+" with "+
			 pairs[0][1]+" attrname:"+pairs[0][2]);

    for(int i=1;k>0;k--,i++){
      pairs[i][0] = (String)sv.elementAt(k);
      pairs[i][1] = (String)tv.elementAt(k);
      pairs[i][2] = attrList[k+offset];
      System.out.println("Replace "+pairs[i][0]+" with "+
			 pairs[i][1]+"attrname:"+pairs[i][2]);
    }
    return pairs;
  }
  
  

  //Delete entry and all of its children
public void makeCutOfEntry(String LDAPHost,int LDAPPort,String oldEntryDN)
  {
        
    LDAPConnection ld=null;
    
    System.out.println("I am deleting all nodes under :"+oldEntryDN);
    Vector deletedDNs = new Vector();
    try {
      ld = connection(LDAPHost,LDAPPort);
      LDAPSearchConstraints cons = ld.getSearchConstraints();
      LDAPSearchResults res = ld.search(oldEntryDN,LDAPConnection.SCOPE_SUB,
		       "objectclass=*", null,  false,  cons );

      /* Loop on results until finished */
      while ( res.hasMoreElements() ) {
	/* Next directory entry */
	LDAPEntry findEntry = (LDAPEntry)res.nextElement();
	deletedDNs.addElement(findEntry.getDN());
      } 
    } catch( LDAPException e ) {
      System.out.println( "makeCutOfEntry:Error: " + e.toString() );
    }

    try{
      for(int k= deletedDNs.size()-1;k>=0;k--)
    	{
	  //After got all the deleted entries, just delete them
	  String DN = (String)deletedDNs.elementAt(k);
	  System.out.println(DN+" will be  deleted");
	  ld.delete(DN);
	}
    } catch( LDAPException e ) {
      System.out.println( "makeCutOfEntry:Error: " + e.toString() );
    }
    
  }

public void makeCopyOfEntry(String LDAPHost,int LDAPPort,String oldEntryDN,
			    String[][] pairs,int SCOPE)
  {
    String RDN;
    
    System.out.println("makeCopyOfEntry called with:oldEntryDN:"+oldEntryDN);
    
    try {
      LDAPConnection ld = connection(LDAPHost,LDAPPort);
      LDAPSearchConstraints cons = ld.getSearchConstraints();
      LDAPSearchResults res = ld.search(oldEntryDN,SCOPE,
		       "objectclass=*", null,  false,  cons );

      System.out.println("makeCopyOfEntry:oldEntryDN:"+oldEntryDN+":");
      /* Loop on results until finished */
      /* Next directory entry */
      LDAPEntry findEntry = (LDAPEntry)res.nextElement();
      makeCopyOfOneEntry(LDAPHost,LDAPPort,findEntry,pairs);
      while ( res.hasMoreElements() ) {
	/* Next directory entry */
	findEntry = (LDAPEntry)res.nextElement();
	makeCopyOfOneEntry(LDAPHost,LDAPPort,findEntry,pairs);
      }
    }catch( LDAPException e ) {
      System.out.println( "makeCopyOfEntry:Error: " + e.toString() );
    }
  }

public String makeCopyOfDN(String oldDN)
  {
    int colon      = oldDN.indexOf(",");
    String rest    = oldDN.substring(colon+1);
    int equal      = oldDN.indexOf("=");
    String newVal  = "CopyOf "+oldDN.substring(equal+1,colon);
    String attName = oldDN.substring(0,equal);
    String newEntryDN     = attName+"="+newVal+","+rest;
    System.out.println("makeCopyOfDN : "+newEntryDN);
    
    return  newEntryDN ;
  }
 
public String updateDN(String DN ,String [][]pairs)
  {
    int start,end;
    if (pairs == null)
      return DN;
    
    for(int k=0;k<pairs.length;k++)
      {
	start = DN.indexOf(pairs[k][0]);
	if (start < 0)
	  continue;
        end = start+pairs[k][0].length()-1;
	DN = DN.substring(0,start)+pairs[k][1]+DN.substring(end+1);
      }
    return DN;
  }

public void updateOneEntry(String LDAPHost,int LDAPPort,String newRDN,
			   String ENTRYDN, String[][] pairs)
  {  
    int len=0;
    
    System.out.println("updateOneEntry:ENTRYDN:"+ENTRYDN);
    LDAPConnection ld = connection(LDAPHost,LDAPPort);
    try{
      //update its attributes
      if (pairs == null){
	ld.rename(ENTRYDN,newRDN,true);
	return;
      }
      LDAPEntry foundEntry = ld.read(ENTRYDN);
      LDAPAttributeSet attrs = foundEntry.getAttributeSet();
      String []fattName  = new String[pairs.length];
      String []fattValue = new String[pairs.length];
      for(int k=0,j=0;k<pairs.length;k++)
	{
	  String attrName = pairs[k][2];
	  //If this entry doesnt have this attribute ,skip to next one
	  if (attrs.getAttribute(attrName) == null)
	    continue;
	  fattName[j]    =attrName ;
	  fattValue[j++] =pairs[k][1].substring(pairs[k][1].indexOf("=")+1);
	  len++;
	  System.out.println("set Attribute :"+attrName+" to :"
			     +fattValue[j-1]);
	  
	}
      String []attName  = new String[len];
      String []attValue = new String[len];
      for(int k=0;k<len;k++){
	attName[k] = fattName[k];
	attValue[k] = fattValue[k];
      }
    
      //update its attributes
      modifyAttribute(LDAPHost, LDAPPort, ENTRYDN,attName,attValue);
      System.out.println("I rename entry:"+ENTRYDN+" newRDN:"+newRDN);
      //update entry  DN
      ld.rename(ENTRYDN,newRDN,true);
      System.out.println("I renamed entry:"+ENTRYDN);
    }catch(LDAPException e ) {
      System.out.println("updateOneentry(error):"+e.toString());
    }
    
    
  }

public Hashtable putEntrytoHash(LDAPEntry findEntry)
  {
    /* Read all attributes */
    Hashtable attributes = new  Hashtable();
    
    /* Get the attributes of the entry */
    LDAPAttributeSet findAttrs = findEntry.getAttributeSet();
    Enumeration enumAttrs = findAttrs.getAttributes();

    System.out.println("reading entries of RDN="+ findEntry.getDN() );
    /* Loop on attributes */
    while ( enumAttrs.hasMoreElements() ) {
      LDAPAttribute anAttr =
	(LDAPAttribute)enumAttrs.nextElement();
      String attrName = anAttr.getName();
      /* Loop on values for this attribute */
      Enumeration enumVals = anAttr.getStringValues();
      Vector attVals = new Vector();
      while(enumVals.hasMoreElements())
	attVals.addElement(enumVals.nextElement());
      attributes.put(attrName,attVals);
    }
    return attributes;  
  }
  
      

public void makeCopyOfRemoteFile(LDAPEntry newEntry,Hashtable atts )
  {
    
    LDAPAttributeSet aset = newEntry.getAttributeSet();
    if (aset.getAttribute("filename") == null)
      return;
    String DN                 = newEntry.getDN();
    String machine = (String)(aset.getAttribute("machine").
			      getStringValues().nextElement());
    String org     = (String)(aset.getAttribute("organization").
			      getStringValues().nextElement());
    String content = getFileFromEntry(atts);
    String []msg ={DN,content,machine,org};
    saveFileInLDAP(msg);
  }
     
   

public void makeCopyOfOneEntry(String LDAPHost,int LDAPPort,
			      LDAPEntry oldEntry, String[][] pairs)
  {  
    String attrName,attrVal;
    
    String newEntryDN =  updateDN(oldEntry.getDN(),pairs);
    LDAPAttributeSet attrs = oldEntry.getAttributeSet();
    Hashtable oldAttrs = putEntrytoHash(oldEntry);

    for(int k=0;(pairs!=null)&&(k<pairs.length);k++)
      {
	attrName = pairs[k][2];
	//If this entry doesnt have this attribte skip to next one
	if (attrs.getAttribute(attrName) == null)
	  continue;
	
	LDAPAttribute attr = new LDAPAttribute(attrName);
	attrVal = pairs[k][1].substring(pairs[k][1].indexOf("=")+1);
	attr.addValue(attrVal);
	//remove old attribue value
	attrs.remove(attrName);
        //replace with new one
	attrs.add(attr);
      }
    
    LDAPConnection ld = connection(LDAPHost,LDAPPort);
    /* Create an entry with this DN and these attributes */
    LDAPEntry myEntry = new LDAPEntry(newEntryDN, attrs );
    /* Now add the entry to the directory */
    System.out.println("myentry :"+ myEntry);
    
    try{
      ld.add( myEntry );
      System.out.println( "Copy of Entry added:" );
      //add remote file if necessary
      makeCopyOfRemoteFile(myEntry,oldAttrs);
      
    }
    catch( LDAPException e ) {
      if ( e.getLDAPResultCode() == LDAPException.ENTRY_ALREADY_EXISTS )
	System.out.println( "Error(adding new host): Entry already present" );
      else
	System.out.println( "Error(adding new host): " + e.toString() );
    }
  }

public boolean modifyAttribute(String LDAPHost,int LDAPPort,String ENTRYDN,
			    String []attName,String []attValue)
{
  
  boolean errorFlag = false;
  if ((attName == null)||(attName.length == 0))
    return errorFlag;

  LDAPModificationSet mods = new LDAPModificationSet();
  for(int k=0;k<attName.length;k++)
    {
      LDAPAttribute newAtt = new LDAPAttribute(attName[k],attValue[k]);
      mods.add( LDAPModification.REPLACE, newAtt);
    }
  
  LDAPConnection ld = connection(LDAPHost,LDAPPort);
  try {
    /* Now modify the entry in the directory */
    System.out.println("I am modifying entry:"+ENTRYDN);
  
    ld.modify( ENTRYDN, mods );
    System.out.println( "Entry modified"  );
  }
  catch( LDAPException e ) {
    errorFlag = true;
    System.out.println( "Error:modifyAttribute:" + e.toString() );
  }
  if (!errorFlag)
    return errorFlag;
  
  //If we could not find the modified entry, create new entry at LDAP server
  System.out.println("I am adding new file at remote site ");
  try{
    createHostEntry(ld,ENTRYDN);
    createUserEntry(ld,ENTRYDN);
    int fIndex = attName.length-1;
    LDAPEntry myEntry = createRemoteLDAPEntry (ld,ENTRYDN,attValue[fIndex]);
    /* Now add the entry to the directory */
    ld.add( myEntry );
    System.out.println( "Entry added"  );
  }
  catch( LDAPException e ) {
    if ( e.getLDAPResultCode() == LDAPException.ENTRY_ALREADY_EXISTS )
      System.out.println( "Error: Entry already present" );
    else
      System.out.println( "Error: " + e.toString() );
  }
  return errorFlag;
    
}

//DNName will be cn=file3,cn=Tom Haupt,hn=Convex,o=NCSA,o=WebFlow,c=US
public void createHostEntry( LDAPConnection ld,String DNName)
  {
    String hostEntryDN = DNName.substring(DNName.indexOf("hn="));
    boolean found=false;
    System.out.println("Host entry Dn:"+hostEntryDN);

    try {
      LDAPEntry foundEntry = ld.read( hostEntryDN);
      found = true;
    } catch ( LDAPException e ) {
      switch( e.getLDAPResultCode() ) {
      case e.NO_SUCH_OBJECT:
	System.out.println("The specified entry does not exist."+hostEntryDN );
	break;
      case e.LDAP_PARTIAL_RESULTS:
	System.out.println( "Entry served by a different LDAP server." );
	break;
      case e.INSUFFICIENT_ACCESS_RIGHTS:
	System.out.println( "You do not have the access rights to perform this operation." );
	break;
      default:
	System.out.println( "Error number: " + e.getLDAPResultCode() );
	System.out.println( "Could not read the specified entry." );
	break;
      }
    }
    
      
      //If we found hostname entry, do not create new one return
    if (found)
      return;
    
    System.out.println("I amd adding new Host:"+hostEntryDN);
    
    String[] list = LDAPDN.explodeDN( hostEntryDN,false);
    
    LDAPAttributeSet attrs = new LDAPAttributeSet();
    LDAPAttribute attr = new LDAPAttribute( "objectclass" );
    String objectclass_values[] = {"top","organization",
				   "organizationalUnit","WebFlowHost"};

    for(int i = 0; i<objectclass_values.length; i++ ) {
      attr.addValue( objectclass_values[i] );
    }
    attrs.add( attr );
  
    attr = new LDAPAttribute( "hostname" );
    String hostname = list[0].substring(list[0].indexOf("=")+1);
    attr.addValue( hostname);
    attrs.add( attr ); 

    attr = new LDAPAttribute( "o" );
    String organization = list[1].substring(list[1].indexOf("=")+1);
    attr.addValue(organization);
    attrs.add( attr ); 

    attr = new LDAPAttribute( "ou");
    String ou = "Chemistry";
    attr.addValue(ou);
    attrs.add( attr ); 
 
    /* Create an entry with this DN and these attributes */
    LDAPEntry myEntry = new LDAPEntry(hostEntryDN, attrs );
    /* Now add the entry to the directory */
    System.out.println("myentry :"+ myEntry);
    
    try{
      ld.add( myEntry );
      System.out.println( "Host Entry added"  );
    }
    catch( LDAPException e ) {
      if ( e.getLDAPResultCode() == LDAPException.ENTRY_ALREADY_EXISTS )
	System.out.println( "Error(adding new host): Entry already present" );
      else
	System.out.println( "Error(adding new host): " + e.toString() );
    }
    
  }
  
    
  
public void createUserEntry( LDAPConnection ld,String DNName)
  {
    String userEntryDN = DNName.substring(DNName.lastIndexOf("cn="));
    boolean found=false;
    System.out.println("User entry dn:"+userEntryDN);
    
    try {
      LDAPEntry foundEntry = ld.read( userEntryDN);
      found = true;
    } catch ( LDAPException e ) {
      switch( e.getLDAPResultCode() ) {
      case e.NO_SUCH_OBJECT:
	System.out.println( "The specified entry does not exist."
			    +userEntryDN );
	break;
      case e.LDAP_PARTIAL_RESULTS:
	System.out.println( "Entry served by a different LDAP server." );
	break;
      case e.INSUFFICIENT_ACCESS_RIGHTS:
	System.out.println( "You do not have the access rights to perform this operation." );
	break;
      default:
	System.out.println( "Error number: " + e.getLDAPResultCode() );
	System.out.println( "Could not read the specified entry." );
	break;
      }
    }
    
      //If we found hostname entry, do not create new one return
    if (found)
      return;
    System.out.println("I amd adding new User:"+userEntryDN);
    String[] list = LDAPDN.explodeDN( userEntryDN,false);
    
    LDAPAttributeSet attrs = new LDAPAttributeSet();
    LDAPAttribute attr = new LDAPAttribute( "objectclass" );
    String objectclass_values[] = {"top","person"};

    for(int i = 0; i<objectclass_values.length; i++ ) {
      attr.addValue( objectclass_values[i] );
    }
    attrs.add( attr );
  
    attr = new LDAPAttribute( "cn" );
    String cn = list[0].substring(list[0].indexOf("=")+1);
    attr.addValue( cn);
    attrs.add( attr );

    attr = new LDAPAttribute( "sn" );
    String sn = cn.substring(cn.indexOf(" ")+1);
    attr.addValue( sn);
    attrs.add( attr );

    /* Create an entry with this DN and these attributes */
    LDAPEntry myEntry = new LDAPEntry(userEntryDN, attrs );
    /* Now add the entry to the directory */
    System.out.println("myentry :"+ myEntry);
    try{
      ld.add( myEntry );
      System.out.println( "User Entry added"  );
    }
    catch( LDAPException e ) {
      if ( e.getLDAPResultCode() == LDAPException.ENTRY_ALREADY_EXISTS )
	System.out.println( "Error(adding new user): Entry already present" );
      else
	System.out.println( "Error(adding new user): " + e.toString() );
    }
  }
  


//DNName will be cn=file3,cn=Tom Haupt,hn=Convex,o=NCSA,o=WebFlow,c=US	 
public LDAPEntry createRemoteLDAPEntry( LDAPConnection ld,String DNName,
					String content)
  {
    String[] list = LDAPDN.explodeDN(DNName,false);
    
    LDAPAttributeSet attrs = new LDAPAttributeSet();
    LDAPAttribute attr = new LDAPAttribute( "objectclass" );
    String objectclass_values[] = {"top","WebFlowFile","person",
				   "WebFlowHost","organization" };

    for(int i = 0; i<objectclass_values.length; i++ ) {
      attr.addValue( objectclass_values[i] );
    }
    attrs.add( attr );

    attr = new LDAPAttribute( "filename" );
    String filename = list[0].substring(list[0].indexOf("=")+1);
    attr.addValue( filename);
    attrs.add( attr );

    attr = new LDAPAttribute( "cn" );
    String cn = list[1].substring(list[1].indexOf("=")+1);
    attr.addValue( cn);
    attrs.add( attr );

    attr = new LDAPAttribute( "sn" );
    String sn = cn.substring(cn.indexOf(" ")+1);
    attr.addValue( sn);
    attrs.add( attr );

    attr = new LDAPAttribute( "hostname" );
    String hostname = list[2].substring(list[2].indexOf("=")+1);
    attr.addValue( hostname);
    attrs.add( attr ); 

    attr = new LDAPAttribute( "o" );
    String organization = list[3].substring(list[3].indexOf("=")+1);
    attr.addValue(organization);
    attrs.add( attr ); 

    attr = new LDAPAttribute( "content" );
    attr.addValue( content);
    attrs.add( attr );

    /* Create an entry with this DN and these attributes */
    LDAPEntry myEntry = new LDAPEntry(DNName, attrs );
    return myEntry;
    
  }
	

public void saveFileInLDAP(Object msg)
  {
    String []params      = (String[])msg;
    String DN            = params[0];
    String fileContent   = params[1];
    String machine       = params[2];
    String organization  = params[3];
   
    System.out.println("DN:"+DN+" fileContent:"+fileContent+
		       " machine:"+machine+" organization:"+organization);
    String filename=DN.substring(DN.indexOf("=")+1,DN.indexOf(","));
  		          
    Hashtable enumAttrs = readEntry(centralLDAPHost,centralLDAPPort,DN) ;
    //replace the values of machine & organization attributes with
    // the incoming values from WFE

    enumAttrs.remove("machine");
    Vector v = new Vector();
    v.addElement(machine);
    enumAttrs.put("machine",v);

    enumAttrs.remove("organization");
    v = new Vector();
    v.addElement(organization);
    enumAttrs.put("organization",v);

    enumAttrs.remove("filename");
    v = new Vector();
    v.addElement(filename);
    enumAttrs.put("filename",v);

    String    remoteLDAPHost = (String)LDAPHostForSite.get(organization);
    if (remoteLDAPHost.equalsIgnoreCase("localhost")){
      System.out.println("Storing into local file");
      //replace the local entry
      String []attrNames = {"organization","machine","filename","content" };
      String []attrValues = {organization,machine, filename,fileContent };
      modifyAttribute(centralLDAPHost,centralLDAPPort,DN,
		      attrNames,attrValues);
    }   
    else{
      System.out.println("remoteLDAPHost :"+remoteLDAPHost);
      
      int index = remoteLDAPHost.indexOf(':');
      int remoteLDAPPort =Integer.parseInt(remoteLDAPHost.substring(index+1));
      remoteLDAPHost = remoteLDAPHost.substring(0,index);
      String remoteRDN = createRDN(enumAttrs);
      //replace remote file
      String []attrNames = {"content" };
      String []attrValues = {fileContent};
      boolean addNew=modifyAttribute(remoteLDAPHost,remoteLDAPPort,remoteRDN,
		      attrNames,attrValues);

      //If we added remote file in remote LDAP server,
      //create new entry in local LDAP server which points to remote file
      if (addNew){
	String []lattrNames = {"organization","machine","filename","content"};
	String []lattrValues = {organization,machine, filename,fileContent};
	modifyAttribute(centralLDAPHost,centralLDAPPort,DN,
		      lattrNames,lattrValues);
	
      }
      
    }
  }
 

public Object recvWFEMessage(int k)
  {
    Object msg=null;
    //msg includes command + ARGS
    try{
      msg = ois_list[k].readObject();
    }catch(OptionalDataException e){
      System.out.println("recvWFEMessage "+e);
    }catch(ClassNotFoundException e){
      System.out.println("recvWFEMessage "+e);
    }catch(IOException e){
      System.out.println("recvWFEMessage "+e);
    }
    return msg;
        
  }
  
 
 
public void  processWFECommand(int k )
  {
    Object msg=null;
    int command = 0;
    String oldEntryDN ;
    String[] params;
    
    //msg includes command + ARGS
    try{
      command = ois_list[k].readInt();
      msg = ois_list[k].readObject();
    }catch(OptionalDataException e){
      System.out.println("processWFECommand "+e);
    }catch(ClassNotFoundException e){
      System.out.println("processWFECommand "+e);
    }catch(IOException e){
      System.out.println("processWFECommand "+e);
    }
    System.out.println("Processing LDAP request :"+command);
    
    switch(command){
    case SAVE:
      saveFileInLDAP(msg);
      sendWFECommandResult(oos_list[k],"OK");
      break;
      

    case OPEN:
      String reqRDN = (String)msg;
      Hashtable enumAttrs=readEntry(centralLDAPHost,centralLDAPPort,reqRDN) ;
      String content=getFileFromEntry(enumAttrs);
      sendWFECommandResult(oos_list[k],content);
      break;

    case DELETE:
       oldEntryDN = (String)msg;
       makeCutOfEntry(centralLDAPHost,centralLDAPPort,oldEntryDN);
      break;
      
    case PASTE:
      params = (String[])msg;
      String source = params[0];
      String target = params[1];
      String editedValue = params[2];
      String prevCmnd = params[3];
      oldEntryDN = params[0];
      System.out.println("Paste called with:source"+source+"target:"+target
			 +" editedValue:"+editedValue+" prevcmnd: "+prevCmnd);

      boolean isCopy = false;
      if (prevCmnd.equals("copy"))
	isCopy = true;
      String[][] pairs = findUnmatchedRDNs(source,target,isCopy,editedValue);
      makeCopyOfEntry(centralLDAPHost,centralLDAPPort,oldEntryDN,pairs,
		     LDAPConnection.SCOPE_SUB );
      if (prevCmnd.equals("cut"))
	makeCutOfEntry(centralLDAPHost,centralLDAPPort,oldEntryDN);
      else if (prevCmnd.equals("copy"))
	System.out.println("Do nothing other that copying original ");
      
      break;

    case NEWFILE:
      params = (String[])msg;
      source = params[0];
      target = params[1];
      editedValue = params[2];
      System.out.println("NEWFILE called with:source"+source+"target:"+target
			 +" editedValue:"+editedValue);
      pairs = findUnmatchedRDNs(source,target,true,editedValue);
      makeCopyOfEntry(centralLDAPHost,centralLDAPPort,source,pairs,
		     LDAPConnection.SCOPE_BASE );
      break;

    case NEWFOLDER:
      oldEntryDN = (String)msg;
      makeCopyOfEntry(centralLDAPHost,centralLDAPPort,oldEntryDN,null,
		      LDAPConnection.SCOPE_BASE);
      break;

    case RENAME:
      params = (String[])msg;
      oldEntryDN = params[0];
      String newRDN     = params[1];
      //String[] childrenDNs = (String[])recvWFEMessage(k);
      //renameEntry(centralLDAPHost,centralLDAPPort, oldEntryDN,
      //	  newRDN,childrenDNs);
      renameEntry(centralLDAPHost,centralLDAPPort, oldEntryDN,
		  newRDN);
      
      break;

    case GETTREE:
      String base = (String)msg;
      searchFullandSend(oos_list[k],base,"objectclass=*");
      break;
      
      
    }
  }


}

